/* Definitions of IRC message functions and list of messages.
 *
 * Epona (c) 2000 PegSoft
 * Contact us at epona@pegirc.com
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include "services.h"
#include "messages.h"
#include "language.h"

static char *uplink;

/* List of messages is at the bottom of the file. */

/*************************************************************************/
/*************************************************************************/

static void m_nickcoll(char *source, int ac, char **av)
{
    if (ac < 1)
	return;
    if (!skeleton && !readonly)
	introduce_user(av[0]);
}

/*************************************************************************/

static void m_ping(char *source, int ac, char **av)
{
    if (ac < 1)
	return;
    send_cmd(ServerName, "PONG %s %s", ac>1 ? av[1] : ServerName, av[0]);
}

/*************************************************************************/

static void m_away(char *source, int ac, char **av)
{
    User *u = finduser(source);

    if (u && (ac == 0 || *av[0] == 0))	/* un-away */
	check_memos(u);
}

/*************************************************************************/

static void m_join(char *source, int ac, char **av)
{
    if (ac != 1)
	return;
    do_join(source, ac, av);
}

/*************************************************************************/

static void m_kick(char *source, int ac, char **av)
{
    if (ac != 3)
	return;
    do_kick(source, ac, av);
}

/*************************************************************************/

static void m_kill(char *source, int ac, char **av)
{
    BotInfo *bi;
	
    if (ac != 2)
	return;
    /* Recover if someone kills us. */
    if (stricmp(av[0], s_OperServ) == 0 ||
        stricmp(av[0], s_NickServ) == 0 ||
        stricmp(av[0], s_ChanServ) == 0 ||
        stricmp(av[0], s_MemoServ) == 0 ||
        (s_BotServ && stricmp(av[0], s_BotServ) == 0) ||
        stricmp(av[0], s_HelpServ) == 0 ||
        (s_DevNull && stricmp(av[0], s_DevNull) == 0) ||
        stricmp(av[0], s_GlobalNoticer) == 0
    ) {
	if (!readonly && !skeleton)
	    introduce_user(av[0]);
    } else if (s_BotServ && (bi = findbot(av[0]))) { 
        if (!readonly && !skeleton)
        {
	    introduce_user(av[0]);
	    bot_rejoin_all(bi);
	}
    } else {
	do_kill(source, ac, av);
    }
}

/*************************************************************************/

static void m_mode(char *source, int ac, char **av)
{
    if (*av[0] == '#' || *av[0] == '&') {
	if (ac < 2)
	    return;
	do_cmode(source, ac, av);
    } else {
	if (ac != 2)
	    return;
	do_umode(source, ac, av);
    }
}

/*************************************************************************/

static void m_motd(char *source, int ac, char **av)
{
    FILE *f;
    char buf[BUFSIZE];

    f = fopen(MOTDFilename, "r");
    send_cmd(ServerName, "375 %s :- %s Message of the Day",
		source, ServerName);
   if (f) {
	while (fgets(buf, sizeof(buf), f)) {
	    buf[strlen(buf)-1] = 0;
	    send_cmd(ServerName, "372 %s :- %s", source, buf);
	}
	fclose(f);
    } else {
	send_cmd(ServerName, "372 %s :- MOTD file not found!  Please "
			"contact your IRC administrator.", source);
    }
    send_cmd(ServerName, "376 %s :End of /MOTD command.", source);
}

/*************************************************************************/

#ifdef IRC_BAHAMUT
static void m_nick(char *source, int ac, char **av)
{
    int i;
    char *av2[2];

    if (ac == 9) {
	av[7] = av[8];
	ac--;
    }
    
    if ((!*source && ac != 8) || (*source && ac != 2)) {
	if (debug) {
	    log("debug: NICK message: expecting 8 or 2 parameters after "
	        "parsing; got %d, source=`%s'", ac, source);
	}
	return;
    }
    
    if (!*source)
    {
    	av2[1] = av[3];
    	for (i = 3; i < 7; i++)
    	    av[i] = av[i+1];	
    	ac--;
    }
    
    do_nick(source, ac, av);
    
    if (!*source) 
    {
    	av2[0] = av[0];
    	do_umode(av[0], 2, av2);
    }
}
#else
static void m_nick(char *source, int ac, char **av)
{
    if (ac == 8) {
	/* Get rid of the useless extra parameter. */
	av[6] = av[7];
	ac--;
    }
    if ((!*source && ac != 7) || (*source && ac != 2)) {
	if (debug) {
	    log("debug: NICK message: expecting 2 or 7 parameters after "
	        "parsing; got %d, source=`%s'", ac, source);
	}
	return;
    }
    do_nick(source, ac, av);
}
#endif

/*************************************************************************/

static void m_part(char *source, int ac, char **av)
{
    if (ac < 1 || ac > 2)
	return;
    do_part(source, ac, av);
}

/*************************************************************************/

static void m_privmsg(char *source, int ac, char **av)
{
    time_t starttime, stoptime;	/* When processing started and finished */
    char *s;
    
    BotInfo *bi;
    ChannelInfo *ci;

    if (ac != 2)
	return;

    /* Check if we should ignore.  Operators always get through. */
    if (allow_ignore && !is_oper(source)) {
	IgnoreData *ign = get_ignore(source);
	if (ign && ign->time > time(NULL)) {
	    log("Ignored message from %s: \"%s\"", source, inbuf);
	    return;
	}
    }

    /* If a server is specified (nick@server format), make sure it matches
     * us, and strip it off. */
    s = strchr(av[0], '@');
    if (s) {
	*s++ = 0;
	if (stricmp(s, ServerName) != 0)
	    return;
    }

    starttime = time(NULL);

    if (stricmp(av[0], s_OperServ) == 0) {
	if (is_oper(source)) {
	    operserv(source, av[1]);
	} else {
	    User *u = finduser(source);
	    if (u)
		notice_lang(s_OperServ, u, ACCESS_DENIED);
	    else
		notice(s_OperServ, source, "Access denied.");
	    if (WallBadOS)
		wallops(s_OperServ, "Denied access to %s from %s (non-oper)",
			s_OperServ, source);
	}
    } else if (stricmp(av[0], s_NickServ) == 0) {
	nickserv(source, av[1]);
    } else if (stricmp(av[0], s_ChanServ) == 0) {
	chanserv(source, av[1]);
    } else if (stricmp(av[0], s_MemoServ) == 0) {
	memoserv(source, av[1]);
    } else if (s_BotServ && stricmp(av[0], s_BotServ) == 0) {
	botserv(source, av[1]);
    } else if (stricmp(av[0], s_HelpServ) == 0) {
	User *u = finduser(source);
	if (u) notice_help(s_HelpServ, u, HELP_HELP, s_NickServ, s_ChanServ, s_MemoServ);
    } else if (s_BotServ && (bi = findbot(av[0]))) {
    	botmsgs(bi, source, av[1]);
    } else if (s_BotServ && (ci = cs_findchan(av[0]))) {
    	if (!(ci->flags & CI_VERBOTEN) && ci->c && ci->bi) /* Some paranoia checks */
    		botchanmsgs(source, ci, av[1]);	
    }

    /* Add to ignore list if the command took a significant amount of time. */
    if (allow_ignore) {
	stoptime = time(NULL);
	if (stoptime > starttime && *source && !strchr(source, '.'))
	    add_ignore(source, stoptime-starttime);
    }
}

/*************************************************************************/

static void m_quit(char *source, int ac, char **av)
{
    if (ac != 1)
	return;
    do_quit(source, ac, av);
}

/*************************************************************************/

#ifdef IRC_UNREAL

static void m_rping(char *source, int ac, char **av)
{
	if (ac < 5) return;
	send_cmd(NULL, "RPONG %s %s %s %s :%s", source, av[1], av[2], av[3], av[4]);
}

#endif

/*************************************************************************/

static void m_server(char *source, int ac, char **av)
{	
	if (!stricmp(av[1], "1"))
		uplink = sstrdup(av[0]);
}

/*************************************************************************/

#ifdef IRC_UNREAL

static void m_sethost(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
	return;
    
    u = finduser(source);
	if (!u) {
	    log("user: SETHOST for nonexistent user %s", source);
	    return;
	}
	
	change_user_host(u, av[0]);
}

/*************************************************************************/

static void m_setident(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
	return;
    
    u = finduser(source);
	if (!u) {
	    log("user: SETIDENT for nonexistent user %s", source);
	    return;
	}
	
	change_user_username(u, av[0]);
}

/*************************************************************************/

static void m_setname(char *source, int ac, char **av)
{
    User *u;
    
    if (ac != 1)
	return;
    
    u = finduser(source);
	if (!u) {
	    log("user: SETNAME for nonexistent user %s", source);
	    return;
	}
	
	change_user_realname(u, av[0]);
}

#endif

/*************************************************************************/

#ifdef IRC_BAHAMUT

static void m_sjoin(char *source, int ac, char **av)
{
	do_sjoin(source, ac, av);
}

#endif

/*************************************************************************/

static void m_stats(char *source, int ac, char **av)
{
	int i;
	NickInfo *ni;
	
    if (ac < 1)
	return;
    switch (*av[0]) {
      case 'c':
      case 'C':
      		send_cmd(NULL, "213 %s N *@%s * %s 0 0", source, RemoteServer, uplink);
      		send_cmd(NULL, "213 %s C *@%s * %s %d 0", source, RemoteServer, uplink, RemotePort);
      		send_cmd(NULL, "219 %s %c :End of /STATS report.", source, *av[0]);
    		break;
    	
      case 'l':
	send_cmd(NULL, "211 %s Server SendBuf SentBytes SentMsgs RecvBuf "
		"RecvBytes RecvMsgs ConnTime", source);
	send_cmd(NULL, "211 %s %s %d %d %d %d %d %d %ld", source, RemoteServer,
		read_buffer_len(), total_read, -1,
		write_buffer_len(), total_written, -1,
		start_time);
	send_cmd(NULL, "219 %s l :End of /STATS report.", source);
	break;
	
	  case 'o':
	  case 'O':		
    		for (i = 0; i < RootNumber; i++)
    			send_cmd(NULL, "243 %s O * * %s Root 0", source, ServicesRoots[i]);
			for (i = 0; i < servadmins.count && (ni = servadmins.list[i]); i++) 
				send_cmd(NULL, "243 %s O * * %s Admin 0", source, ni->nick);
			for (i = 0; i < servopers.count && (ni = servopers.list[i]); i++) 
				send_cmd(NULL, "243 %s O * * %s Oper 0", source, ni->nick);
					
			send_cmd(NULL, "219 %s %c :End of /STATS report.", source, *av[0]);
	break;
	
      case 'u': {
	int uptime = time(NULL) - start_time;
	send_cmd(NULL, "242 %s :Services up %d day%s, %02d:%02d:%02d",
		source, uptime/86400, (uptime/86400 == 1) ? "" : "s",
		(uptime/3600) % 24, (uptime/60) % 60, uptime % 60);
	send_cmd(NULL, "250 %s :Current users: %d (%d ops); maximum %d",
		source, usercnt, opcnt, maxusercnt);
	send_cmd(NULL, "219 %s u :End of /STATS report.", source);
	break;
      } /* case 'u' */

      case 'h':
      case 'i':
      case 'k':
      case 'm':
      case 'y':
	send_cmd(NULL, "219 %s %c :/STATS %c not applicable or not supported.",
		source, *av[0], *av[0]);
	break;
    }
}

/*************************************************************************/

static void m_time(char *source, int ac, char **av)
{
    time_t t;
    struct tm *tm;
    char buf[64];

    time(&t);
    tm = localtime(&t);
    strftime(buf, sizeof(buf), "%a %b %d %H:%M:%S %Y %Z", tm);
    send_cmd(NULL, "391 :%s", buf);
}

/*************************************************************************/

static void m_topic(char *source, int ac, char **av)
{
    if (ac != 4)
	return;
    do_topic(source, ac, av);
}

/*************************************************************************/

void m_version(char *source, int ac, char **av)
{
    if (source)
	send_cmd(ServerName, "351 %s Epona-%s %s :-- %s",
			source, version_number, ServerName, version_build);
}

/*************************************************************************/

void m_whois(char *source, int ac, char **av)
{
    BotInfo *bi;
    const char *clientdesc;

    if (source && ac >= 1) {
	if (stricmp(av[0], s_NickServ) == 0)
	    clientdesc = desc_NickServ;
	else if (stricmp(av[0], s_ChanServ) == 0)
	    clientdesc = desc_ChanServ;
	else if (stricmp(av[0], s_MemoServ) == 0)
	    clientdesc = desc_MemoServ;
	else if (s_BotServ && stricmp(av[0], s_BotServ) == 0)
	    clientdesc = desc_BotServ;
	else if (stricmp(av[0], s_HelpServ) == 0)
	    clientdesc = desc_HelpServ;
	else if (stricmp(av[0], s_OperServ) == 0)
	    clientdesc = desc_OperServ;
	else if (stricmp(av[0], s_GlobalNoticer) == 0)
	    clientdesc = desc_GlobalNoticer;
	else if (s_DevNull && stricmp(av[0], s_DevNull) == 0)
	    clientdesc = desc_DevNull;
	else if (s_BotServ && (bi = findbot(av[0]))) {
	    /* Bots are handled separately */
	    send_cmd(ServerName, "311 %s %s %s %s * :%s", source, bi->nick, bi->user, bi->host, bi->real);
	    send_cmd(ServerName, "307 %s :%s is a registered nick", source, bi->nick);
	    send_cmd(ServerName, "312 %s %s %s :%s", source, bi->nick, ServerName, ServerDesc);
	    send_cmd(ServerName, "317 %s %s %ld %ld :seconds idle, signon time", source, bi->nick, time(NULL) - bi->lastmsg, start_time);
	    send_cmd(ServerName, "318 %s %s :End of /WHOIS list.", source, bi->nick);
	    return;
	}
	else {
	    send_cmd(ServerName, "401 %s %s :No such service.", source, av[0]);
	    return;
	}
	send_cmd(ServerName, "311 %s %s %s %s * :%s", source, av[0], ServiceUser, ServiceHost, clientdesc);
	send_cmd(ServerName, "312 %s %s %s :%s", source, av[0], ServerName, ServerDesc);
	send_cmd(ServerName, "317 %s %s %ld %ld :seconds idle, signon time", source, av[0], time(NULL) - start_time, start_time);
	send_cmd(ServerName, "318 %s %s :End of /WHOIS list.", source, av[0]);
    }
}

/*************************************************************************/
/*************************************************************************/

Message messages[] = {

    { "401",       NULL },
    { "436",       m_nickcoll },
    { "AWAY",      m_away },
    { "INVITE",    NULL },
    { "JOIN",      m_join },
    { "KICK",      m_kick },
    { "KILL",      m_kill },
    { "MODE",      m_mode },
    { "MOTD",      m_motd },
    { "NICK",      m_nick },
    { "NOTICE",    NULL },
    { "PART",      m_part },
    { "PASS",      NULL },
    { "PING",      m_ping },
    { "PRIVMSG",   m_privmsg },
    { "QUIT",      m_quit },
    { "SERVER",    m_server },
    { "SQUIT",     NULL },
    { "STATS",     m_stats },
    { "TIME",      m_time },
    { "TOPIC",     m_topic },
    { "USER",      NULL },
    { "VERSION",   m_version },
    { "WALLOPS",   NULL },
    { "WHOIS",     m_whois },

    /* DALnet specific messages */
    { "AKILL",     NULL },
    { "GLOBOPS",   NULL },
    { "GNOTICE",   NULL },
    { "GOPER",     NULL },
    { "RAKILL",    NULL },
    { "SILENCE",   NULL },
    { "SQLINE",    NULL },
    { "UNSQLINE",  NULL },
    
    /* DreamForge specific messages */
#ifdef IRC_DAL4_4_15
    { "PROTOCTL",  NULL },
#endif
    
    /* Bahamut specific messages */
#ifdef IRC_BAHAMUT
    { "CAPAB", 	   NULL },
    { "SJOIN",     m_sjoin },
    { "SVINFO",    NULL },
#endif

#ifdef IRC_UNREAL
	{ "LAG",       NULL },
	{ "NETINFO",   NULL },
	{ "RPING",     m_rping },
	{ "SETHOST",   m_sethost },
	{ "SETIDENT",  m_setident },
	{ "SETNAME",   m_setname },
#endif

    { NULL }

};

/*************************************************************************/

Message *find_message(const char *name)
{
    Message *m;

    for (m = messages; m->name; m++) {
	if (stricmp(name, m->name) == 0)
	    return m;
    }
    return NULL;
}

/*************************************************************************/
